/*
 * Copyright © 2009 Université Bordeaux
 * Copyright © 2020-2022 Inria.  All rights reserved.
 *
 * See COPYING in top-level directory.
 */

#ifndef HWLOC_PRIVATE_WINDOWS_H
#define HWLOC_PRIVATE_WINDOWS_H

#ifndef _ANONYMOUS_UNION
#ifdef __GNUC__
#define _ANONYMOUS_UNION __extension__
#else
#define _ANONYMOUS_UNION
#endif /* __GNUC__ */
#endif /* _ANONYMOUS_UNION */

#ifndef _ANONYMOUS_STRUCT
#ifdef __GNUC__
#define _ANONYMOUS_STRUCT __extension__
#else
#define _ANONYMOUS_STRUCT
#endif /* __GNUC__ */
#endif /* _ANONYMOUS_STRUCT */

#define DUMMYUNIONNAME
#define DUMMYSTRUCTNAME

#endif /* HWLOC_PRIVATE_WINDOWS_H */
