/*
 * Copyright (C) by Argonne National Laboratory
 *     See COPYRIGHT in top-level directory
 */

#ifndef STUBSHM_PRE_H_INCLUDED
#define STUBSHM_PRE_H_INCLUDED

typedef struct {
    int dummy;
} MPIDI_STUBSHM_request_t;

typedef struct {
    int dummy;
} MPIDI_STUBSHM_am_request_t;

typedef struct MPIDI_STUBSHM_comm_t {
    int dummy;
} MPIDI_STUBSHM_comm_t;

#endif /* STUBSHM_PRE_H_INCLUDED */
