/*
 * Copyright (C) by Argonne National Laboratory
 *     See COPYRIGHT in top-level directory
 */

#ifndef SHM_IPC_INLINE_H_INCLUDED
#define SHM_IPC_INLINE_H_INCLUDED

#include "ipc_send.h"

/* Not-inlined shm functions */
#include "ipc_noinline.h"

#endif /* SHM_IPC_INLINE_H_INCLUDED */
