/*
 * Copyright (C) by Argonne National Laboratory
 *     See COPYRIGHT in top-level directory
 */

#include <mpiimpl.h>
#include <mpir_typerep.h>

void MPIR_Typerep_init(void)
{
    return;
}

void MPIR_Typerep_finalize(void)
{
    return;
}
