/*
 * Copyright (C) by Argonne National Laboratory
 *     See COPYRIGHT in top-level directory
 */

#include "mpitest.h"

/* This file provides initialization/finalization functions for single-threaded
 * tests. */

void MTest_init_thread_pkg(void)
{
}

void MTest_finalize_thread_pkg(void)
{
}
