/*
 * Copyright (C) by Argonne National Laboratory
 *     See COPYRIGHT in top-level directory
 */

#ifndef SHM_AM_FALLBACK_H_INCLUDED
#define SHM_AM_FALLBACK_H_INCLUDED

#include "shm_am_fallback_send.h"
#include "shm_am_fallback_recv.h"
#include "shm_am_fallback_probe.h"
#include "shm_am_fallback_rma.h"
#include "shm_am_fallback_coll.h"
#include "shm_am_fallback_part.h"

#endif /* SHM_AM_FALLBACK_H_INCLUDED */
