/*
 * Copyright (C) by Argonne National Laboratory
 *     See COPYRIGHT in top-level directory
 */

#ifndef POSIX_EAGER_STUB_IMPL_H_INCLUDED
#define POSIX_EAGER_STUB_IMPL_H_INCLUDED

#include <mpidimpl.h>
#include <posix_eager_transaction.h>

#endif /* POSIX_EAGER_STUB_IMPL_H_INCLUDED */
